# pyproject2setup.py -- setuptools support
# vim:se fileencoding=utf-8 :
# (c) 2020 Eli Schwartz
# 2-clause BSD license

from __future__ import absolute_import

import os
import subprocess
import sys


def handle_setuptools(data):
    """
    Handle pyproject.toml unserialized into data, by ignoring it and using the
    setuptools build system instead.

    Prefer running the contents of setup.py, but fall back to running a setup()
    function.
    """
    # TODO: shouldn't we be ignoring it with non-legacy backend?
    if os.path.exists('setup.py'):
        ret = (subprocess.Popen([sys.executable, 'setup.py'] + sys.argv[1:])
               .wait())
        if ret != 0:
            sys.exit(ret)
    else:
        from setuptools import setup
        setup()


def get_handlers():
    """
    Return build-backend mapping for setuptools.
    """

    return {'setuptools.build_meta': handle_setuptools,
            'setuptools.build_meta:__legacy__': handle_setuptools}



# pyproject2setup.py -- PEP 621 metadata support
# vim:se fileencoding=utf-8 :
# (c) 2021 Michał Górny
# 2-clause BSD license

from collections import defaultdict


def get_pep621_metadata(data, allow_dynamic=[]):
    """
    Get PEP 621 metadata if available, return None otherwise.
    """

    if 'project' not in data:
        return None
    metadata = data['project']

    # TODO: follow PEP621 author/email split?
    authors = []
    author_emails = []
    for a in metadata['authors']:
        if 'name' in a:
            authors.append(a['name'])
        if 'email' in a:
            author_emails.append(a['email'])

    entry_points = defaultdict(list)
    if 'scripts' in metadata:
        for name, content in metadata['scripts'].items():
            entry_points['console_scripts'].append(
                '{} = {}'.format(name, content)
            )

    if 'gui-scripts' in metadata:
        for name, content in metadata['gui-scripts'].items():
            entry_points['gui_scripts'].append(
                '{} = {}'.format(name, content)
            )

    if 'entrypoints' in metadata:
        for group_name, group_content in metadata['entrypoints'].items():
            if group_name in ('console_scripts', 'gui_scripts'):
                raise ValueError('{} forbidden in entrypoints'
                                 .format(group_name))
            for name, path in group_content.items():
                entry_points[group_name].append(
                    '{} = {}'.format(name, path)
                )

    for key in allow_dynamic:
        has_static = key in metadata
        has_dynamic = key in metadata.get('dynamic', [])
        if has_static and has_dynamic:
            raise ValueError('Key {} declared both statically and as dynamic'
                             .format(key))
        if not has_static and not has_dynamic:
            raise ValueError('Key {} must be declared either statically or as '
                             'dynamic'.format(key))

    return {
        'name': metadata['name'],
        'version': metadata.get('version'),
        'description': metadata.get('description'),
        'author': ', '.join(authors),
        'author_email': ', '.join(author_emails),
        'classifiers': metadata.get('classifiers', []),
        'entry_points': dict(entry_points),
    }



# pyproject2setup.py -- main handler
# vim:se fileencoding=utf-8 :
# (c) 2019-2020 Michał Górny
# 2-clause BSD license

from __future__ import absolute_import

try:
    import tomli as toml
    OPEN_FLAGS = 'rb'
except ImportError:
    import toml
    OPEN_FLAGS = 'r'

import pyproject2setuppy.flit
import pyproject2setuppy.poetry
import pyproject2setuppy.setuptools


MODULES = (
    pyproject2setuppy.flit,
    pyproject2setuppy.poetry,
    pyproject2setuppy.setuptools,
)


def get_handlers():
    """
    Get mapping of build-backend values for supported build systems.
    """

    handlers = {}
    for m in MODULES:
        handlers.update(m.get_handlers())
    return handlers


def main():
    """
    Run setuptools' setup() function for pyproject.toml in the current
    working directory.
    """

    with open('pyproject.toml', OPEN_FLAGS) as f:
        data = toml.load(f)
    backend = data['build-system']['build-backend']

    handler = get_handlers().get(backend)
    if handler is None:
        raise NotImplementedError(
                'Build backend {} unknown'.format(backend))

    handler(data)


if __name__ == '__main__':
    main()



# pyproject2setup.py -- common routines
# vim:se fileencoding=utf-8 :
# (c) 2019-2020 Michał Górny
# 2-clause BSD license

from __future__ import absolute_import

from collections import defaultdict

from setuptools import find_packages

import os.path


def raise_exc(e):
    raise e


def auto_find_packages(modname, subdir='.'):
    """
    Find packages for modname, and supply proper setup() args for them.
    Supports both packages and modules in correct directory.  Includes
    all nested subpackages.
    """
    retdict = {}
    if subdir != '.':
        retdict['package_dir'] = {'': subdir}
    if os.path.isdir(os.path.join(subdir, modname)):
        retdict.update(
            {'packages': find_packages(where=subdir,
                                       include=(modname,
                                                '{}.*'.format(modname)))})
    elif os.path.isfile(os.path.join(subdir, modname + '.py')):
        retdict.update({'py_modules': [modname]})
    else:
        raise RuntimeError('No package matching {} found'.format(modname))
    return retdict


def find_package_data(packages, package_dirs={}):
    """Find additional package data dirs and return package_data dict."""
    ret = defaultdict(list)
    # install all data files from package directories
    ret[''] = ['*']

    # find data subdirectories
    for p in packages:
        pkgdir = package_dirs.get(p, os.path.join(package_dirs.get('', ''),
                                                  p.replace('.', '/')))
        for topdir, dirs, files in os.walk(pkgdir, onerror=raise_exc):
            dirs[:] = [x for x in dirs if x != '__pycache__'
                       and not x.startswith('.')]
            if '__init__.py' not in files:
                data_path = os.path.relpath(topdir, pkgdir)
                ret[p].append(data_path + '/*')

    return dict((x, sorted(frozenset(y))) for (x, y) in ret.items())



# pyproject2setup.py -- backwards compatibility for .main module
# vim:se fileencoding=utf-8 :
# (c) 2021 Michał Górny
# 2-clause BSD license

from __future__ import absolute_import

from pyproject2setuppy.__main__ import get_handlers, main


__all__ = [get_handlers, main]


if __name__ == '__main__':
    main()



# pyproject2setup.py -- poetry support
# vim:se fileencoding=utf-8 :
# (c) 2019-2021 Michał Górny
# 2-clause BSD license

from __future__ import absolute_import

from setuptools import find_packages, setup

from collections import defaultdict

import email.utils
import os.path
import re

from pyproject2setuppy.common import auto_find_packages, find_package_data


def handle_poetry(data):
    """
    Handle pyproject.toml unserialized into data, using poetry build
    system.
    """

    metadata = data['tool']['poetry']

    authors = []
    author_emails = []
    for a in metadata['authors']:
        name, addr = email.utils.parseaddr(a)
        authors.append(name)
        author_emails.append(addr)

    if 'packages' not in metadata:
        # canonicalize the name
        canonical_name = re.sub(r'[-.]', '_', metadata['name'].lower())
        try:
            package_args = auto_find_packages(canonical_name)
        except RuntimeError:
            package_args = auto_find_packages(canonical_name, 'src')
    else:
        package_args = {'packages': [], 'package_dir': {}}
        for p in metadata['packages']:
            if p.get('format', '') == 'sdist':
                continue
            subdir = p.get('from', '.')
            packages = find_packages(
                subdir, include=(p['include'], p['include'] + '.*'))
            package_args['packages'].extend(packages)
            if subdir != '.':
                for sp in packages:
                    package_args['package_dir'][sp] = os.path.join(
                        subdir, sp.replace('.', os.path.sep))

    package_args['package_data'] = (
        find_package_data(package_args.get('packages', []),
                          package_args.get('package_dir', {})))

    # NB: include doesn't seem to do anything without exclude
    if metadata.get('exclude', []):
        raise NotImplementedError('exclude is not implemented yet')

    entry_points = defaultdict(list)
    if 'scripts' in metadata:
        for name, content in metadata['scripts'].items():
            entry_points['console_scripts'].append(
                '{} = {}'.format(name, content)
            )

    if 'plugins' in metadata:
        for group_name, group_content in metadata['plugins'].items():
            for name, path in group_content.items():
                entry_points[group_name].append(
                    '{} = {}'.format(name, path)
                )

    setup(name=metadata['name'],
          version=metadata['version'],
          description=metadata['description'],
          author=', '.join(authors),
          author_email=', '.join(author_emails),
          url=metadata.get('homepage'),
          classifiers=metadata.get('classifiers', []),
          entry_points=dict(entry_points),
          **package_args)


def get_handlers():
    """
    Return build-backend mapping for poetry.
    """

    return {'poetry.masonry.api': handle_poetry,
            'poetry.core.masonry.api': handle_poetry,
            }



# pyproject2setup.py -- flit support
# vim:se fileencoding=utf-8 :
# (c) 2019-2021 Michał Górny
# 2-clause BSD license

from __future__ import absolute_import

from setuptools import setup

from collections import defaultdict

import importlib
import os.path
import sys

from pyproject2setuppy.common import auto_find_packages, find_package_data
from pyproject2setuppy.pep621 import get_pep621_metadata


def handle_flit(data):
    """
    Handle pyproject.toml unserialized into data, using flit build
    system.
    """

    # try PEP 621 first
    setup_metadata = get_pep621_metadata(data, ['version', 'description'])
    modname = None
    if setup_metadata is not None:
        if 'metadata' in data.get('tool', {}).get('flit', {}):
            raise ValueError('[project] and [tool.flit.metadata] cannot be '
                             'present simultaneously')

        # see if tool.flit.module specifies another module name
        modname = (data.get('tool', {}).get('flit', {})
                   .get('module', {}).get('name', None))
    else:
        # tool.flit fallback
        topdata = data['tool']['flit']
        metadata = topdata['metadata']

        entry_points = defaultdict(list)
        if 'scripts' in topdata:
            for name, content in topdata['scripts'].items():
                entry_points['console_scripts'].append(
                    '{} = {}'.format(name, content)
                )

        if 'entrypoints' in topdata:
            for group_name, group_content in topdata['entrypoints'].items():
                for name, path in group_content.items():
                    entry_points[group_name].append(
                        '{} = {}'.format(name, path)
                    )

        setup_metadata = {
            'name': metadata['module'],
            # use None to match PEP 621 return value for dynamic
            'version': None,
            'description': None,
            'author': metadata['author'],
            'author_email': metadata['author-email'],
            'url': metadata.get('home-page'),
            'classifiers': metadata.get('classifiers', []),
            'entry_points': dict(entry_points),
        }

    # handle dynamic metadata if necessary
    if modname is None:
        modname = setup_metadata['name']

    if None in [setup_metadata[x] for x in ('version', 'description')]:
        sys.path.insert(0, '.')
        sys.path.insert(1, 'src')
        mod = importlib.import_module(modname.replace('/', '.'), '')
        if setup_metadata['version'] is None:
            setup_metadata['version'] = mod.__version__
        if setup_metadata['description'] is None:
            # setuptools doesn't like multiple lines in description
            setup_metadata['description'] = (
                ' '.join(mod.__doc__.strip().splitlines()))

    try:
        setup_metadata.update(auto_find_packages(modname))
    except RuntimeError:
        setup_metadata.update(auto_find_packages(modname, 'src'))
    setup_metadata['package_data'] = (
        find_package_data(setup_metadata.get('packages', []),
                          setup_metadata.get('package_dir', {})))

    setup(**setup_metadata)


def handle_flit_thyself(data):
    """Handle flit_core.build_thyself backend"""
    bs = data['build-system']
    backend_path = bs['backend-path']
    if not isinstance(backend_path, list):
        backend_path = [backend_path]
    sys.path = backend_path + sys.path
    mod = importlib.import_module(bs['build-backend'], '')
    metadata = mod.metadata_dict
    package_args = auto_find_packages(bs['build-backend'].split('.')[0])

    setup(name=mod.metadata.name,
          version=mod.metadata.version,
          description=mod.metadata.summary,
          author=metadata['author'],
          author_email=metadata['author_email'],
          url=metadata.get('home_page'),
          classifiers=metadata.get('classifiers', []),
          **package_args)


def get_handlers():
    """
    Return build-backend mapping for flit.
    """

    return {'flit.buildapi': handle_flit,
            'flit_core.buildapi': handle_flit,
            'flit_core.build_thyself': handle_flit_thyself,
            }
